#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.common import TORRENT_STATE

log = logging.getLogger(__name__)

STATE_SORT = ['All', 'Active'] + TORRENT_STATE


# Special purpose filters:
def filter_keywords(torrent_ids, values):
    # Cleanup
    keywords = ','.join([v.lower() for v in values])
    keywords = keywords.split(',')

    for keyword in keywords:
        torrent_ids = filter_one_keyword(torrent_ids, keyword)
    return torrent_ids


def filter_one_keyword(torrent_ids, keyword):
    """
    search torrent on keyword.
    searches title,state,tracker-status,tracker,files
    """
    all_torrents = component.get('TorrentManager').torrents

    for torrent_id in torrent_ids:
        torrent = all_torrents[torrent_id]
        if keyword in torrent.filename.lower():
            yield torrent_id
        elif keyword in torrent.state.lower():
            yield torrent_id
        elif torrent.trackers and keyword in torrent.trackers[0]['url']:
            yield torrent_id
        elif keyword in torrent_id:
            yield torrent_id
        # Want to find broken torrents (search on "error", or "unregistered")
        elif keyword in torrent.tracker_status.lower():
            yield torrent_id
        else:
            for t_file in torrent.get_files():
                if keyword in t_file['path'].lower():
                    yield torrent_id
                    break


def filter_by_name(torrent_ids, search_string):
    all_torrents = component.get('TorrentManager').torrents
    try:
        search_string, match_case = search_string[0].split('::match')
    except ValueError:
        search_string = search_string[0]
        match_case = False

    if match_case is False:
        search_string = search_string.lower()

    for torrent_id in torrent_ids:
        torrent_name = all_torrents[torrent_id].get_name()
        if match_case is False:
            torrent_name = all_torrents[torrent_id].get_name().lower()
        else:
            torrent_name = all_torrents[torrent_id].get_name()

        if search_string in torrent_name:
            yield torrent_id


def tracker_error_filter(torrent_ids, values):
    filtered_torrent_ids = []
    tm = component.get('TorrentManager')

    # If this is a tracker_host, then we need to filter on it
    if values[0] != 'Error':
        for torrent_id in torrent_ids:
            if values[0] == tm[torrent_id].get_status(['tracker_host'])['tracker_host']:
                filtered_torrent_ids.append(torrent_id)
        return filtered_torrent_ids

    # Check torrent's tracker_status for 'Error:' and return those torrent_ids
    for torrent_id in torrent_ids:
        if 'Error:' in tm[torrent_id].get_status(['tracker_status'])['tracker_status']:
            filtered_torrent_ids.append(torrent_id)
    return filtered_torrent_ids


class FilterManager(component.Component):
    """FilterManager"""

    def __init__(self, core):
        component.Component.__init__(self, 'FilterManager')
        log.debug('FilterManager init..')
        self.core = core
        self.torrents = core.torrentmanager
        self.registered_filters = {}
        self.register_filter('keyword', filter_keywords)
        self.register_filter('name', filter_by_name)
        self.tree_fields = {}

        self.register_tree_field('state', self._init_state_tree)

        def _init_tracker_tree():
            return {'Error': 0}

        self.register_tree_field('tracker_host', _init_tracker_tree)

        self.register_filter('tracker_host', tracker_error_filter)

        def _init_users_tree():
            return {'': 0}

        self.register_tree_field('owner', _init_users_tree)

    def filter_torrent_ids(self, filter_dict):
        """
        returns a list of torrent_id's matching filter_dict.
        core filter method
        """
        if not filter_dict:
            return self.torrents.get_torrent_list()

        # Sanitize input: filter-value must be a list of strings
        for key, value in filter_dict.items():
            if isinstance(value, str):
                filter_dict[key] = [value]

        # Optimized filter for id
        if 'id' in filter_dict:
            torrent_ids = list(filter_dict['id'])
            del filter_dict['id']
        else:
            torrent_ids = self.torrents.get_torrent_list()

        # Return if there's nothing more to filter
        if not filter_dict:
            return torrent_ids

        # Special purpose, state=Active.
        if 'state' in filter_dict:
            # We need to make sure this is a list for the logic below
            filter_dict['state'] = list(filter_dict['state'])

        if 'state' in filter_dict and 'Active' in filter_dict['state']:
            filter_dict['state'].remove('Active')
            if not filter_dict['state']:
                del filter_dict['state']
            torrent_ids = self.filter_state_active(torrent_ids)

        if not filter_dict:
            return torrent_ids

        # Registered filters
        for field, values in list(filter_dict.items()):
            if field in self.registered_filters:
                # Filters out doubles
                torrent_ids = list(
                    set(self.registered_filters[field](torrent_ids, values))
                )
                del filter_dict[field]

        if not filter_dict:
            return torrent_ids

        torrent_keys, plugin_keys = self.torrents.separate_keys(
            list(filter_dict), torrent_ids
        )
        # Leftover filter arguments, default filter on status fields.
        for torrent_id in list(torrent_ids):
            status = self.core.create_torrent_status(
                torrent_id, torrent_keys, plugin_keys
            )
            for field, values in filter_dict.items():
                if field in status and status[field] in values:
                    continue
                elif torrent_id in torrent_ids:
                    torrent_ids.remove(torrent_id)
        return torrent_ids

    def get_filter_tree(self, show_zero_hits=True, hide_cat=None):
        """
        returns {field: [(value,count)] }
        for use in sidebar.
        """
        torrent_ids = self.torrents.get_torrent_list()
        tree_keys = list(self.tree_fields)
        if hide_cat:
            for cat in hide_cat:
                tree_keys.remove(cat)

        torrent_keys, plugin_keys = self.torrents.separate_keys(tree_keys, torrent_ids)
        items = {field: self.tree_fields[field]() for field in tree_keys}

        for torrent_id in list(torrent_ids):
            status = self.core.create_torrent_status(
                torrent_id, torrent_keys, plugin_keys
            )  # status={key:value}
            for field in tree_keys:
                value = status[field]
                items[field][value] = items[field].get(value, 0) + 1

        if 'tracker_host' in items:
            items['tracker_host']['All'] = len(torrent_ids)
            items['tracker_host']['Error'] = len(
                tracker_error_filter(torrent_ids, ('Error',))
            )

        if not show_zero_hits:
            for cat in ['state', 'owner', 'tracker_host']:
                if cat in tree_keys:
                    self._hide_state_items(items[cat])

        # Return a dict of tuples:
        sorted_items = {field: sorted(items[field].items()) for field in tree_keys}

        if 'state' in tree_keys:
            sorted_items['state'].sort(key=self._sort_state_item)

        return sorted_items

    def _init_state_tree(self):
        init_state = {}
        init_state['All'] = len(self.torrents.get_torrent_list())
        for state in TORRENT_STATE:
            init_state[state] = 0
        init_state['Active'] = len(
            self.filter_state_active(self.torrents.get_torrent_list())
        )
        return init_state

    def register_filter(self, filter_id, filter_func, filter_value=None):
        self.registered_filters[filter_id] = filter_func

    def deregister_filter(self, filter_id):
        del self.registered_filters[filter_id]

    def register_tree_field(self, field, init_func=lambda: {}):
        self.tree_fields[field] = init_func

    def deregister_tree_field(self, field):
        if field in self.tree_fields:
            del self.tree_fields[field]

    def filter_state_active(self, torrent_ids):
        for torrent_id in list(torrent_ids):
            status = self.torrents[torrent_id].get_status(
                ['download_payload_rate', 'upload_payload_rate']
            )
            if status['download_payload_rate'] or status['upload_payload_rate']:
                pass
            else:
                torrent_ids.remove(torrent_id)
        return torrent_ids

    def _hide_state_items(self, state_items):
        """For hide(show)-zero hits"""
        for value, count in list(state_items.items()):
            if value != 'All' and count == 0:
                del state_items[value]

    def _sort_state_item(self, item):
        try:
            return STATE_SORT.index(item[0])
        except ValueError:
            return 99



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#


"""PluginManagerBase"""

import email
import importlib.abc
import importlib.machinery
import logging
import sys
import unicodedata
import zipfile
from functools import cached_property
from importlib.metadata import EntryPoints
from importlib.resources import files
from pathlib import Path

from twisted.internet import defer
from twisted.python.failure import Failure

import deluge.component as component
import deluge.configmanager
from deluge.common import VersionSplit

log = logging.getLogger(__name__)

METADATA_KEYS = [
    'Name',
    'License',
    'Author',
    'Home-page',
    'Summary',
    'Platform',
    'Version',
    'Author-email',
    'Description',
]

DEPRECATION_WARNING = """
The plugin %s is not using the "deluge_" namespace.
In order to avoid package name clashes between regular python packages and
deluge plugins, the way deluge plugins should be created has changed.
If you're seeing this message and you're not the developer of the plugin which
triggered this warning, please report to it's author.
If you're the developer, please take a look at the plugins hosted on deluge's
git repository to have an idea of what needs to be changed.
"""


def _extract_egg(egg_path: Path, dest: Path) -> None:
    """Extract a .egg zip to dest, skipping EGG-INFO metadata."""
    dest.mkdir(parents=True, exist_ok=True)
    with zipfile.ZipFile(egg_path) as zf:
        for member in zf.namelist():
            if not member.startswith('EGG-INFO/'):
                zf.extract(member, dest)


def _read_egg_info(egg_info: 'Path | zipfile.Path') -> tuple[str, str]:
    """Read PKG-INFO and entry_points.txt from an egg-info directory.

    Accepts a filesystem Path or zipfile.Path pointing directly at the
    egg-info directory (e.g. EGG-INFO/ or PluginName.egg-info/).
    """
    pkg_info = (egg_info / 'PKG-INFO').read_text()
    ep_path = egg_info / 'entry_points.txt'
    ep_txt = ep_path.read_text() if ep_path.is_file() else ''
    return pkg_info, ep_txt


def _ep_module_root(entry_points: 'EntryPoints') -> str | None:
    """Return the top-level package name from the first entry point."""
    if ep := next(iter(entry_points), None):
        return ep.value.split(':')[0].split('.')[0]
    return None


def plugin_name_to_id(name: str) -> str:
    """Convert a plugin display name to its canonical id (lowercase kebab-case).

    Applies NFKD unicode normalization and ASCII encoding to handle non-ASCII names.
    Matches Python distribution name convention (PEP 503).
    """
    ascii_name = unicodedata.normalize('NFKD', name).encode('ascii', 'ignore').decode()
    return ascii_name.lower().replace(' ', '-').replace('_', '-')


def _load_plugin_metadata(
    item_path: Path,
) -> tuple[Path, str, str] | None:
    """Detect plugin type and read (location, pkg_info, ep_txt), or None."""
    match item_path.suffix:
        case '.egg' if item_path.is_file():
            location = item_path
            pkg_info, ep_txt = _read_egg_info(zipfile.Path(location, 'EGG-INFO'))
        case '.egg' if item_path.is_dir():
            location = item_path
            pkg_info, ep_txt = _read_egg_info(location / 'EGG-INFO')
        case '.egg-info' if item_path.is_dir():
            location = item_path.parent
            pkg_info, ep_txt = _read_egg_info(item_path)
        case '.egg-link' if item_path.is_file():
            location = Path(item_path.read_text().splitlines()[0].strip())
            pkg_info, ep_txt = _read_egg_info(next(location.glob('*.egg-info')))
        case _:
            return None
    return location, pkg_info, ep_txt


def _parse_pkg_info(pkg_info: str) -> dict:
    """Parse a PKG-INFO string and return a dict of metadata fields."""
    metadata_msg = email.message_from_string(pkg_info)
    metadata_ver = metadata_msg.get('Metadata-Version')

    info = {key: metadata_msg.get(key, '') for key in METADATA_KEYS}

    # Optional Description field in body (Metadata spec >=2.1)
    if not info['Description'] and (metadata_ver or '').startswith('2'):
        info['Description'] = metadata_msg.get_payload().strip()

    return info


class _PluginDirFinder(importlib.abc.MetaPathFinder):
    """MetaPathFinder that loads top-level packages from a single directory.

    Used for both directory plugins and extracted zip egg plugins. Registered
    on sys.meta_path so it can be cleanly removed on plugin deactivation,
    without sys.path pollution or module cache issues.
    """

    def __init__(self, directory: Path):
        self._finder = importlib.machinery.FileFinder(
            str(directory),
            (importlib.machinery.SourceFileLoader, ['.py']),
            (importlib.machinery.SourcelessFileLoader, ['.pyc']),
        )

    def find_spec(self, fullname, _path, target=None):
        """Find a module spec for a top-level plugin package.

        Only handles top-level imports (e.g. deluge_foo), not namespace
        subpackage imports (e.g. deluge.plugins.foo).
        """
        if '.' in fullname:
            return None
        return self._finder.find_spec(fullname, target)

    def invalidate_caches(self):
        self._finder.invalidate_caches()


class _PluginDist:
    """Discovered plugin distribution.

    Attributes:
        name: Display name from PKG-INFO (e.g. "Foo Bar").
        id: Normalised distribution id (e.g. "foo-bar"), PEP 503.
        module_root: Top-level package from entry point (e.g. "deluge_foobar").
    """

    def __init__(
        self,
        name: str,
        version: str,
        location: Path,
        entry_points: EntryPoints,
        pkg_info: str,
    ):
        self.name = name
        self.id = plugin_name_to_id(name)
        self.version = version
        self.location = location
        self._entry_points = entry_points
        self._pkg_info = pkg_info
        self._finder: _PluginDirFinder | None = None
        self._active: bool = False
        self.module_root: str = _ep_module_root(entry_points) or self.id.replace(
            '-', '_'
        )

    @property
    def is_active(self) -> bool:
        """Returns True if the plugin is loaded and active, False if not."""
        return self._active

    @cached_property
    def info(self) -> dict:
        """Return a dict of plugin metadata fields."""
        return _parse_pkg_info(self._pkg_info)

    @property
    def entry_points(self) -> EntryPoints:
        """Return all the plugin's entry points."""
        return self._entry_points

    @cached_property
    def _egg_cache_dir(self) -> Path:
        cache_base = Path(deluge.configmanager.get_config_dir()) / 'plugin_cache'
        return cache_base / f'{self.id}-{self.version}'

    @classmethod
    def from_metadata(cls, location: Path, pkg_info: str, ep_txt: str) -> '_PluginDist':
        meta = email.message_from_string(pkg_info)
        entry_points = EntryPoints._from_text_for(ep_txt, None)
        return cls(
            name=meta.get('Name', ''),
            version=meta.get('Version', ''),
            location=location,
            entry_points=entry_points,
            pkg_info=pkg_info,
        )

    def activate(self):
        if self.is_active:
            return

        is_plugin_egg = self.location.is_file()

        if is_plugin_egg:
            plugin_dir = self._egg_cache_dir
            # Extract if new plugin egg or egg has changed
            if not plugin_dir.exists() or (
                self.location.stat().st_mtime_ns > plugin_dir.stat().st_mtime_ns
            ):
                _extract_egg(self.location, plugin_dir)
        else:
            plugin_dir = self.location

        self._finder = _PluginDirFinder(plugin_dir)
        sys.meta_path.insert(0, self._finder)
        self._active = True

    def deactivate(self):
        if self._finder:
            try:
                sys.meta_path.remove(self._finder)
            except ValueError:
                pass
            self._finder = None
        # Evict cached modules so a subsequent activate() re-imports cleanly.
        for key in list(sys.modules):
            if key == self.module_root or key.startswith(self.module_root + '.'):
                del sys.modules[key]
        self._active = False


class PluginManagerBase:
    """PluginManagerBase is a base class for PluginManagers to inherit"""

    def __init__(
        self,
        config_file: str,
        entry_name: str,
        plugin_dirs: list[str | Path] | None = None,
    ) -> None:
        """Initialise the plugin manager.

        Args:
            config_file: Name of the config file (e.g. 'core.conf').
            entry_name: The setuptools entry-point group to load plugins from
                (e.g. 'deluge.plugin.core').
            plugin_dirs: Directories to scan for plugins. Defaults to None,
                which uses the standard base and user plugin directories.
                Primarily intended for testing.
        """
        log.debug('Plugin manager init..')

        self.config = deluge.configmanager.ConfigManager(config_file)

        # Create the plugins folder if it doesn't exist
        (Path(deluge.configmanager.get_config_dir()) / 'plugins').mkdir(exist_ok=True)

        # This is the entry we want to load..
        self.entry_name = entry_name

        # Loaded plugins
        self.plugins = {}

        # Directories scanned for plugins
        self.plugin_dirs = plugin_dirs if plugin_dirs else self.default_plugin_dirs()

        # Scan the plugin folders for plugins
        self.scan_for_plugins()

    def enable_plugins(self):
        # Load plugins that are enabled in the config.
        for name in self.config['enabled_plugins']:
            self.enable_plugin(name)

    def disable_plugins(self):
        """Disable all plugins that are enabled"""
        # Dict will be modified so iterate over generated list
        for key in list(self.plugins):
            self.disable_plugin(key)

    def __getitem__(self, key):
        return self.plugins[key]

    def get_available_plugins(self):
        """Returns a list of the available plugins name"""
        return self.available_plugins

    def get_enabled_plugins(self):
        """Returns a list of enabled plugins"""
        return list(self.plugins)

    @staticmethod
    def default_plugin_dirs() -> list[Path]:
        """Returns the default directories to scan for plugins."""
        base_dir = Path(str(files('deluge') / 'plugins'))
        user_dir = Path(deluge.configmanager.get_config_dir()) / 'plugins'
        base_subdir = [item for item in base_dir.iterdir() if item.is_dir()]
        return [base_dir, user_dir] + base_subdir

    def scan_for_plugins(self) -> None:
        """Scan plugin_dirs for available plugins."""
        self._plugin_dist = {}
        self.available_plugins = []

        for scan_dir in self.plugin_dirs:
            scan_path = Path(scan_dir)
            if not scan_path.is_dir():
                continue
            for item in scan_path.iterdir():
                self.register_plugin(item)

    def register_plugin(self, plugin_path: Path) -> None:
        try:
            plugin_metadata = _load_plugin_metadata(plugin_path)
            plugin_dist = (
                _PluginDist.from_metadata(*plugin_metadata) if plugin_metadata else None
            )
        except Exception:
            log.warning(
                'Skipping %s: could not read plugin metadata',
                plugin_path,
                exc_info=True,
            )
            return

        if plugin_dist is None or not plugin_dist.name:
            return

        log.debug(
            'Found plugin: %s %s at %s',
            plugin_dist.name,
            plugin_dist.version,
            plugin_dist.location,
        )
        key = plugin_dist.id
        existing = self._plugin_dist.get(key)
        if existing:
            new_ver = VersionSplit(plugin_dist.version)
            old_ver = VersionSplit(existing.version)
            # Prefer newer version, or a source directory over a plugin of equal version.
            prefer_new = new_ver > old_ver or (
                new_ver == old_ver
                and existing.location.is_file()
                and plugin_dist.location.is_dir()
            )
            if not prefer_new:
                log.debug('Skipping duplicate plugin: %s at %s', key, plugin_path)
                return
            log.debug(
                'Replacing plugin %s %s at %s with %s at %s',
                key,
                existing.version,
                existing.location,
                plugin_dist.version,
                plugin_dist.location,
            )
            self.available_plugins.remove(existing.name)
        self._plugin_dist[key] = plugin_dist
        self.available_plugins.append(plugin_dist.name)

    def enable_plugin(self, plugin_name):
        """Enable a plugin.

        Args:
            plugin_name (str): The plugin name.

        Returns:
            Deferred: A deferred with callback value True or False indicating
                whether the plugin is enabled or not.

        """
        if plugin_name not in self.available_plugins:
            log.warning('Cannot enable non-existent plugin %s', plugin_name)
            return defer.succeed(False)

        if plugin_name in self.plugins:
            log.warning('Cannot enable already enabled plugin %s', plugin_name)
            return defer.succeed(True)

        plugin_key = plugin_name_to_id(plugin_name)
        plugin_dist = self._plugin_dist[plugin_key]
        plugin_dist.activate()
        return_d = defer.succeed(True)

        for ep in plugin_dist.entry_points.select(group=self.entry_name):
            try:
                cls = ep.load()
                instance = cls(plugin_dist.id)
            except component.ComponentAlreadyRegistered as ex:
                log.error(ex)
                return defer.succeed(False)
            except Exception as ex:
                log.error(
                    'Unable to instantiate plugin %r from %r!',
                    ep.name,
                    plugin_dist.location,
                )
                log.exception(ex)
                continue
            try:
                return_d = defer.maybeDeferred(instance.enable)
            except Exception as ex:
                log.error('Unable to enable plugin: %s', ep.name)
                log.exception(ex)
                return_d = defer.fail(False)

            if not instance.__module__.startswith('deluge_'):
                import warnings

                warnings.warn_explicit(
                    DEPRECATION_WARNING % ep.name,
                    DeprecationWarning,
                    instance.__module__,
                    0,
                )
            if self._component_state == 'Started':

                def on_enabled(result, instance):
                    return component.start([instance.plugin._component_name])

                return_d.addCallback(on_enabled, instance)

            def on_started(result, instance):
                display_name = plugin_dist.name
                self.plugins[display_name] = instance
                if display_name not in self.config['enabled_plugins']:
                    log.debug(
                        'Adding %s to enabled_plugins list in config', display_name
                    )
                    self.config['enabled_plugins'].append(display_name)
                log.info('Plugin %s enabled...', display_name)
                return True

            def on_started_error(result, instance):
                log.error(
                    'Failed to start plugin: %s\n%s',
                    plugin_dist.name,
                    result.getTraceback(elideFrameworkCode=1, detail='brief'),
                )
                self.plugins[plugin_dist.name] = instance
                self.disable_plugin(plugin_dist.name)
                return False

            return_d.addCallbacks(
                on_started,
                on_started_error,
                callbackArgs=[instance],
                errbackArgs=[instance],
            )
            return return_d

        return defer.succeed(False)

    def disable_plugin(self, name):
        """Disable a plugin.

        Args:
            plugin_name (str): The plugin name.

        Returns:
            Deferred: A deferred with callback value True or False indicating
                whether the plugin is disabled or not.

        """
        if name not in self.plugins:
            log.warning('Plugin "%s" is not enabled...', name)
            return defer.succeed(True)

        try:
            d = defer.maybeDeferred(self.plugins[name].disable)
        except Exception as ex:
            log.error(
                'Error when disabling plugin: %s',
                self.plugins[name].plugin._component_name,
            )
            log.debug(ex)
            d = defer.succeed(False)

        def on_disabled(result):
            ret = True
            if isinstance(result, Failure):
                log.debug(
                    'Error when disabling plugin %s: %s', name, result.getTraceback()
                )
                ret = False
            try:
                component.deregister(self.plugins[name].plugin)
                del self.plugins[name]
                self.config['enabled_plugins'].remove(name)
                plugin_dist = self._plugin_dist.get(plugin_name_to_id(name))
                if plugin_dist:
                    plugin_dist.deactivate()
            except Exception as ex:
                log.warning('Problems occurred disabling plugin: %s', name)
                log.debug(ex)
                ret = False
            else:
                log.info('Plugin %s disabled...', name)
            return ret

        d.addBoth(on_disabled)
        return d

    def get_plugin_info(self, name):
        """Returns a dictionary of plugin info from the metadata"""
        plugin_dist = self._plugin_dist.get(plugin_name_to_id(name))
        if not plugin_dist:
            log.warning('Failed to retrieve info for plugin: %s', name)
            return {
                **dict.fromkeys(METADATA_KEYS, ''),
                'Name': 'not available',
                'Version': 'not available',
                'Location': '',
            }

        return {**plugin_dist.info, 'Location': str(plugin_dist.location)}



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import signal
import struct
import sys
from typing import Tuple

import deluge.component as component
import deluge.ui.console.utils.colors as colors
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.utils.format_utils import remove_formatting

try:
    import curses
    import curses.panel
except ImportError:
    pass

try:
    from fcntl import ioctl
    from termios import TIOCGWINSZ
except ImportError:
    pass


log = logging.getLogger(__name__)


class InputKeyHandler:
    def __init__(self):
        self._input_result = None

    def set_input_result(self, result):
        self._input_result = result

    def get_input_result(self):
        result = self._input_result
        self._input_result = None
        return result

    def handle_read(self, c):
        """Handle a character read from curses screen

        Returns:
            int: One of the constants defined in util.curses_util.ReadState.
            ReadState.IGNORED: The key was not handled. Further processing should continue.
            ReadState.READ: The key was read and processed. Do no further processing
            ReadState.CHANGED: The key was read and processed. Internal state was changed
            leaving data to be read by the caller.

        """
        return util.ReadState.IGNORED


class TermResizeHandler:
    def __init__(self):
        try:
            signal.signal(signal.SIGWINCH, self.on_resize)
        except ValueError as ex:
            log.debug('TermResize unavailable, unable to catch SIGWINCH signal: %s', ex)
        except AttributeError as ex:
            log.debug('TermResize unavailable, no SIGWINCH signal on Windows: %s', ex)

    @staticmethod
    def get_window_size(fd: int = 0) -> Tuple[int, int]:
        """Return the tty window size as row, col."""
        return struct.unpack('4h', ioctl(fd, TIOCGWINSZ, b'\x00' * 8))[0:2]

    def on_resize(self, _signum, _frame):
        """Handler for SIGWINCH when terminal changes size"""
        rows, cols = self.get_window_size()
        curses.resizeterm(rows, cols)
        return rows, cols


class CursesStdIO:
    """
    fake fd to be registered as a reader with the twisted reactor.
       Curses classes needing input should extend this
    """

    def fileno(self):
        """We want to select on FD 0"""
        return 0

    def doRead(self):  # NOQA: N802
        """called when input is ready"""
        pass

    def logPrefix(self):  # NOQA: N802
        return 'CursesClient'


class BaseMode(CursesStdIO, component.Component):
    def __init__(
        self, stdscr, encoding=None, do_refresh=True, mode_name=None, depend=None
    ):
        """
        A mode that provides a curses screen designed to run as a reader in a twisted reactor.
        This mode doesn't do much, just shows status bars and "Base Mode" on the screen

        Modes should subclass this and provide overrides for:

        do_read(self) - Handle user input
        refresh(self) - draw the mode to the screen
        add_string(self, row, string) - add a string of text to be displayed.
        see method for detailed info

        The init method of a subclass *must* call BaseMode.__init__

        Useful fields after calling BaseMode.__init__:
        self.stdscr - the curses screen
        self.rows - # of rows on the curses screen
        self.cols - # of cols on the curses screen
        self.topbar - top statusbar
        self.bottombar - bottom statusbar
        """
        self.mode_name = mode_name if mode_name else self.__class__.__name__
        component.Component.__init__(self, self.mode_name, 1, depend=depend)
        self.stdscr = stdscr
        # Make the input calls non-blocking
        self.stdscr.nodelay(1)

        self.paused = False
        # Strings for the 2 status bars
        self.statusbars = component.get('StatusBars')
        self.help_hstr = '{!status!} Press {!magenta,blue,bold!}[h]{!status!} for help'

        # Keep track of the screen size
        self.rows, self.cols = self.stdscr.getmaxyx()

        if not encoding:
            self.encoding = sys.getdefaultencoding()
        else:
            self.encoding = encoding

        # Do a refresh right away to draw the screen
        if do_refresh:
            self.refresh()

    def on_resize(self, rows, cols):
        self.rows, self.cols = rows, cols

    def connectionLost(self, reason):  # NOQA: N802
        self.close()

    def add_string(self, row, string, scr=None, **kwargs):
        if scr:
            screen = scr
        else:
            screen = self.stdscr

        return add_string(row, string, screen, self.encoding, **kwargs)

    def draw_statusbars(
        self,
        top_row=0,
        bottom_row=-1,
        topbar=None,
        bottombar=None,
        bottombar_help=True,
        scr=None,
    ):
        self.add_string(top_row, topbar if topbar else self.statusbars.topbar, scr=scr)
        bottombar = bottombar if bottombar else self.statusbars.bottombar
        if bottombar_help:
            if bottombar_help is True:
                bottombar_help = self.help_hstr
            bottombar += (
                ' '
                * (
                    self.cols
                    - len(remove_formatting(bottombar))
                    - len(remove_formatting(bottombar_help))
                )
                + bottombar_help
            )
        self.add_string(self.rows + bottom_row, bottombar, scr=scr)

    # This mode doesn't do anything with popups
    def set_popup(self, popup):
        pass

    def pause(self):
        self.paused = True

    def mode_paused(self):
        return self.paused

    def resume(self):
        self.paused = False
        self.refresh()

    def refresh(self):
        """
        Refreshes the screen.
        Updates the lines based on the`:attr:lines` based on the `:attr:display_lines_offset`
        attribute and the status bars.
        """
        self.stdscr.erase()
        self.draw_statusbars()
        # Update the status bars

        self.add_string(1, '{!info!}Base Mode (or subclass has not overridden refresh)')

        self.stdscr.redrawwin()
        self.stdscr.refresh()

    def doRead(self):  # NOQA: N802
        """
        Called when there is data to be read, ie, input from the keyboard.
        """
        # We wrap this function to catch exceptions and shutdown the mainloop
        try:
            self.read_input()
        except Exception as ex:  # pylint: disable=broad-except
            log.exception(ex)

    def read_input(self):
        # Read the character
        self.stdscr.getch()
        self.stdscr.refresh()

    def close(self):
        """
        Clean up the curses stuff on exit.
        """
        curses.nocbreak()
        self.stdscr.keypad(0)
        curses.echo()


def add_string(
    row, fstring, screen, encoding, col=0, pad=True, pad_char=' ', trim='..', leaveok=0
):
    """
    Adds a string to the desired `:param:row`.

    Args:
        row(int): the row number to write the string
        row(int): the row number to write the string
        fstring(str): the (formatted) string of text to add
        scr(curses.window): optional window to add string to instead of self.stdscr
        col(int): optional starting column offset
        pad(bool): optional bool if the string should be padded out to the width of the screen
        trim(bool): optional bool if the string should be trimmed if it is too wide for the screen

    The text can be formatted with color using the following format:

    "{!fg, bg, attributes, ...!}"

    See: http://docs.python.org/library/curses.html#constants for attributes.

    Alternatively, it can use some built-in scheme for coloring.
    See colors.py for built-in schemes.

    "{!scheme!}"

    Examples:

    "{!blue, black, bold!}My Text is {!white, black!}cool"
    "{!info!}I am some info text!"
    "{!error!}Uh oh!"

    Returns:
        int: the next row

    """
    try:
        parsed = colors.parse_color_string(fstring)
    except colors.BadColorString as ex:
        log.error('Cannot add bad color string %s: %s', fstring, ex)
        return

    if leaveok:
        screen.leaveok(leaveok)

    max_y, max_x = screen.getmaxyx()
    for index, (color, string) in enumerate(parsed):
        # Skip printing chars beyond max_x
        if col >= max_x:
            break

        if index + 1 == len(parsed) and pad:
            # This is the last string so lets append some padding to it
            string += pad_char * (max_x - (col + len(string)))

        if col + len(string) > max_x:
            remaining_chrs = max(0, max_x - col)
            if trim:
                string = string[0 : max(0, remaining_chrs - len(trim))] + trim
            else:
                string = string[0:remaining_chrs]

        try:
            screen.addstr(row, col, string.encode(encoding), color)
        except curses.error:
            # Ignore exception for writing offscreen.
            pass

        col += len(string)

    if leaveok:
        screen.leaveok(0)

    return row + 1


def mkpanel(color, rows, cols, tly, tlx):
    win = curses.newwin(rows, cols, tly, tlx)
    pan = curses.panel.new_panel(win)
    if curses.has_colors():
        win.bkgdset(ord(' '), curses.color_pair(color))
    else:
        win.bkgdset(ord(' '), curses.A_BOLD)
    return pan


def mkwin(color, rows, cols, tly, tlx):
    win = curses.newwin(rows, cols, tly, tlx)
    if curses.has_colors():
        win.bkgdset(ord(' '), curses.color_pair(color))
    else:
        win.bkgdset(ord(' '), curses.A_BOLD)
    return win


def mkpad(color, rows, cols):
    win = curses.newpad(rows, cols)
    if curses.has_colors():
        win.bkgdset(ord(' '), curses.color_pair(color))
    else:
        win.bkgdset(ord(' '), curses.A_BOLD)
    return win


def move_cursor(screen, row, col):
    try:
        screen.move(row, col)
    except curses.error as ex:
        import traceback

        log.warning(
            'Error on screen.move(%s, %s): (curses.LINES: %s, curses.COLS: %s) Error: %s\nStack: %s',
            row,
            col,
            curses.LINES,
            curses.COLS,
            ex,
            ''.join(traceback.format_stack()),
        )



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""

The AlertManager handles all the libtorrent alerts.

This should typically only be used by the Core. Plugins should utilize the
`:mod:EventManager` for similar functionality.

"""

import contextlib
import logging
import threading
import time
from collections import defaultdict
from functools import partial
from typing import Any, Callable

from twisted.internet import reactor, task, threads

import deluge.component as component
from deluge._libtorrent import lt
from deluge.common import decode_bytes

log = logging.getLogger(__name__)


class AlertManager(component.Component):
    """AlertManager fetches and processes libtorrent alerts"""

    def __init__(self):
        log.debug('AlertManager init...')
        component.Component.__init__(self, 'AlertManager')
        self.session = component.get('Core').session

        # Increase the alert queue size so that alerts don't get lost.
        self.alert_queue_size = 10000
        self.set_alert_queue_size(self.alert_queue_size)

        alert_mask = (
            lt.alert.category_t.error_notification
            | lt.alert.category_t.port_mapping_notification
            | lt.alert.category_t.storage_notification
            | lt.alert.category_t.tracker_notification
            | lt.alert.category_t.status_notification
            | lt.alert.category_t.ip_block_notification
            | lt.alert.category_t.performance_warning
            | lt.alert.category_t.file_progress_notification
        )

        self.session.apply_settings({'alert_mask': alert_mask})

        # handlers is a dictionary of lists {"alert_type": [handler1,h2,..]}
        self.handlers = defaultdict(list)
        self.handlers_timeout_secs = 2
        self.delayed_calls = []
        self._event = threading.Event()

    def update(self):
        pass

    def start(self):
        thread = threading.Thread(
            target=self.wait_for_alert_in_thread, name='alert-poller', daemon=True
        )
        thread.start()
        self._event.set()

    def stop(self):
        self.cancel_delayed_calls()

    def pause(self):
        self._event.clear()

    def resume(self):
        self._event.set()

    def wait_for_alert_in_thread(self):
        while self._component_state not in ('Stopping', 'Stopped'):
            if self.check_delayed_calls():
                time.sleep(0.05)
                continue

            if self.session.wait_for_alert(1000) is None:
                continue
            if self._event.wait():
                threads.blockingCallFromThread(reactor, self.maybe_handle_alerts)

    def on_delayed_call_timeout(self, result, timeout, **kwargs):
        log.warning('Alert handler was timed-out before being called %s', kwargs)

    def cancel_delayed_calls(self):
        """Cancel all delayed handlers."""
        for delayed_call in self.delayed_calls:
            delayed_call.cancel()
        self.delayed_calls = []

    def check_delayed_calls(self) -> bool:
        """Returns True if any handler calls are delayed."""
        self.delayed_calls = [dc for dc in self.delayed_calls if not dc.called]
        return len(self.delayed_calls) > 0

    def maybe_handle_alerts(self) -> None:
        if self._component_state != 'Started':
            return

        self.handle_alerts()

    def register_handler(self, alert_type: str, handler: Callable[[Any], None]) -> None:
        """
        Registers a function that will be called when 'alert_type' is pop'd
        in handle_alerts.  The handler function should look like: handler(alert)
        Where 'alert' is the actual alert object from libtorrent.

        Args:
            alert_type: String representation of the libtorrent alert name.
                Can be supplied with or without `_alert` suffix.
            handler: Callback function when the alert is raised.
        """
        if alert_type and alert_type.endswith('_alert'):
            alert_type = alert_type[: -len('_alert')]

        self.handlers[alert_type].append(handler)
        log.debug('Registered handler for alert %s', alert_type)

    def deregister_handler(self, handler: Callable[[Any], None]):
        """
        De-registers the `handler` function from all alert types.

        Args:
            handler: The handler function to deregister.
        """
        for alert_type_handlers in self.handlers.values():
            with contextlib.suppress(ValueError):
                alert_type_handlers.remove(handler)

    def handle_alerts(self):
        """
        Pops all libtorrent alerts in the session queue and handles them appropriately.
        """
        alerts = self.session.pop_alerts()
        if not alerts:
            return

        num_alerts = len(alerts)
        if log.isEnabledFor(logging.DEBUG):
            log.debug('Alerts queued: %s', num_alerts)
        if num_alerts > 0.9 * self.alert_queue_size:
            log.warning(
                'Warning total alerts queued, %s, passes 90%% of queue size.',
                num_alerts,
            )

        for alert in alerts:
            alert_type = alert.what()

            # Display the alert message
            if log.isEnabledFor(logging.DEBUG):
                log.debug('%s: %s', alert_type, decode_bytes(alert.message()))

            if alert_type not in self.handlers:
                continue

            # Call any handlers for this alert type
            for handler in self.handlers[alert_type]:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug('Handling alert: %s', alert_type)
                d = task.deferLater(reactor, 0, handler, alert)
                on_handler_timeout = partial(
                    self.on_delayed_call_timeout,
                    handler=handler.__qualname__,
                    alert_type=alert_type,
                )
                d.addTimeout(
                    self.handlers_timeout_secs,
                    reactor,
                    onTimeoutCancel=on_handler_timeout,
                )
                self.delayed_calls.append(d)

    def set_alert_queue_size(self, queue_size):
        """Sets the maximum size of the libtorrent alert queue"""
        log.info('Alert Queue Size set to %s', queue_size)
        self.alert_queue_size = queue_size
        component.get('Core').apply_session_setting(
            'alert_queue_size', self.alert_queue_size
        )
